import hexer
import io
import pathlib
import re
import tempfile
import unittest

from hexer import regs, main

class test_handlers( unittest.TestCase ):
    def setUp( self ):
        # test file content
        tf = b'\0\0\0\1\0\x0a\0\x0bABCD"FGH\0\2\0\3\0\0\0\0'
        self.data_file = io.BytesIO()
        self.data_file.write( tf )
        self.data_file.seek( 0 )

        self.hx_file = io.StringIO()
        self.out_file = io.StringIO()

        self.hx = hexer.hexer( self.data_file, self.hx_file, self.out_file )

    def assertInOut( self, handler, reg, input, output, *a, **kw ):
        self.assertEqual( list( handler( reg.match( input ), self.hx, *a, **kw ) ), output )

    def test_filepos_token( self ):
        self.assertInOut( hexer.filepos_token, regs.c.filepos, ':', [ '00:' ] )

    def test_filepos_token_bad_pos( self ):
        with self.assertRaises( hexer.FilePosMissmatch ) as cx:
            self.assertInOut( hexer.filepos_token, regs.c.filepos, '100:', [] )
        self.assertEqual( str( cx.exception ), 'file position missmatch 0x0 vs 0x100' )

    def test_filepos_token_skip( self ):
        self.assertInOut( hexer.filepos_token, regs.c.filepos, '--- 4:', [ '--- 04:' ] )
        self.assertEqual( self.hx.data_file.tell(), 4 )

    def test_filepos_token_skip_output( self ):
        with self.assertRaises( NotImplementedError ):
            self.assertInOut( hexer.filepos_token, regs.c.filepos, '*--- 4:', [] )

    def test_filepos_token_rel( self ):
        self.data_file.seek( 4 )
        self.assertInOut( hexer.filepos_token, regs.c.filepos, '+4:', [ '08:' ] )
        self.assertEqual( self.hx.data_file.tell(), 8 )

    def test_comment_token( self ):
        self.assertInOut( hexer.comment_token, regs.c.comment, '# abc', [ '# abc' ] )

    def test_hex_token( self ):
        self.assertInOut( hexer.hex_token, regs.c.hex, '.', [ '00' ] )

    def test_hex_token_assert( self ):
        self.assertInOut( hexer.hex_token, regs.c.hex, '=00000001', [ '=', '00000001' ] )

    def test_hex_token_assert_bad( self ):
        with self.assertRaises( hexer.HexDataAssertError ) as cx:
            self.assertInOut( hexer.hex_token, regs.c.hex, '=00000002', [] )

        self.assertEqual( str( cx.exception ), '00000002 != 00000001' )

    def test_endian_token( self ):
        self.assertInOut( hexer.endian_token, regs.c.endian, '@>', [ '@>' ] )
        self.assertEqual( self.hx.endianess, '>' )

    def test_types_token( self ):
        self.test_endian_token()
        self.assertInOut( hexer.types_token, regs.c.types, 'L', [ 'L[1]' ] )
        self.data_file.seek(0)
        self.assertInOut( hexer.types_token, regs.c.types, 'L{a}', [ 'L{a}[1]' ] )
        self.assertEqual( self.hx._vars['a'], 1 )

    def test_types_token_mult( self ):
        self.test_endian_token()
        self.assertInOut( hexer.types_token, regs.c.types, 'LHH', [ 'LHH[1,10,11]' ] )

    def test_string_token( self ):
        self.data_file.seek( 8 )
        self.assertInOut( hexer.string_token, regs.c.string, 'S*8', [ r'S*8"ABCD\"FGH"' ] )
        self.data_file.seek( 8 )
        self.assertInOut( hexer.string_token, regs.c.string, 'S{name}*8', [ r'S{name}*8"ABCD\"FGH"' ] )
        self.assertEqual( self.hx._vars['name'], b'ABCD"FGH' )

    def test_macro_def_token( self ):
        self.assertInOut( hexer.macro_def_token, regs.c.macrodef, '<=abc>.</abc>', [ '<=abc>.</abc>' ] )
        self.assertEqual( self.hx.macros['abc'], '.' )

    def test_macro_use_token( self ):
        self.test_macro_def_token()
        self.assertInOut( hexer.macro_use_token, regs.c.macrouse, '<abc />', [ '<abc>','00','</abc>' ] )
        self.assertInOut( hexer.macro_use_token, regs.c.macrouse, '<abc ns/>', [ '<abc ns>','00','</abc>' ] )

    def test_macro_not_found( self ):
        with self.assertRaises( hexer.MacroNotFound ) as cx:
            self.assertInOut( hexer.macro_use_token, regs.c.macrouse, '<abc />', [] )

        self.assertEqual( str( cx.exception ), 'abc' )

    def test_macro_import_token( self ):
        with tempfile.NamedTemporaryFile( 'w' ) as tmp:
            tmp.write( '.' )
            tmp.flush()

            self.assertInOut( hexer.macro_import_token, regs.c.macroimport, '<+"' + tmp.name + '"/>', [ '<+"' + tmp.name + '"/>' ] )
            self.assertInOut( hexer.macro_import_token, regs.c.macroimport, '<+"' + tmp.name + '"></+>', [ '<+"' + tmp.name + '">', '00' ,'</+>' ] )

    def test_repeat_token( self ):
        self.test_endian_token()

        token = ( hexer.types_token, regs.c.types.match( 'B' ) )

        self.data_file.seek( 0 )
        self.assertInOut( hexer.repeat_token, regs.c.repeat, '*1', [ 'B[0]' ], token = token )

        self.data_file.seek( 0 )
        self.assertInOut( hexer.repeat_token, regs.c.repeat, '*2', [ 'B[0]', 'B[0]' ], token = token )

        self.data_file.seek( 0 )
        self.assertInOut( hexer.repeat_token, regs.c.repeat, '*3', [ 'B[0]', 'B[0]', 'B[0]' ], token = token )

        self.data_file.seek( 0 )
        self.assertInOut( hexer.repeat_token, regs.c.repeat, ' * 2', [ 'B[0]', ' ', 'B[0]', ' ' ], token = token )

        self.data_file.seek( 0 )
        self.assertInOut( hexer.repeat_token, regs.c.repeat, '*2*3', [ 'B[0]', 'B[0]', 'B[0]', 'B[1]', 'B[0]', 'B[10]' ], token = token )

        self.data_file.seek( 0 )
        self.assertInOut( hexer.repeat_token, regs.c.repeat, ' *2 *3 *4', [
            'B[0]', ' ', 'B[0]', ' ', ' ',
            'B[0]', ' ', 'B[1]', ' ', ' ',
            'B[0]', ' ', 'B[10]', ' ', ' ', ' ',
            'B[0]', ' ', 'B[11]', ' ', ' ',
            'B[65]', ' ', 'B[66]', ' ', ' ',
            'B[67]', ' ', 'B[68]', ' ', ' ', ' ',
            'B[34]', ' ', 'B[70]', ' ', ' ',
            'B[71]', ' ', 'B[72]', ' ', ' ',
            'B[0]', ' ', 'B[2]', ' ', ' ', ' ',
            'B[0]', ' ', 'B[3]', ' ', ' ',
            'B[0]', ' ', 'B[0]', ' ', ' ',
            'B[0]', ' ', 'B[0]', ' ', ' ', ' ' ], token = token )

class test_regs( unittest.TestCase ):
    def test_comment( self ):
        m = regs.c.comment.match( '# abc' )
        self.assertIsInstance( m, re.Match )

    def test_space( self ):
        m = regs.c.space.match( '  ' )
        self.assertIsInstance( m, re.Match )

    def test_filepos( self ):
        m = regs.c.filepos.match( '1234:' )
        self.assertIsInstance( m, re.Match )

    def test_hex( self ):
        m = regs.c.hex.match( 'abc123' )
        self.assertIsInstance( m, re.Match )

    def test_endian( self ):
        m = regs.c.endian.match( '@>' )
        self.assertIsInstance( m, re.Match )

    def test_dtypes( self ):
        m = regs.c.dtypes.match( 'LLHB' )
        self.assertIsInstance( m, re.Match )

    def test_types( self ):
        m = regs.c.types.match( 'LLHB[1,2,3,4]' )
        self.assertIsInstance( m, re.Match )

    def test_string( self ):
        m = regs.c.string.match( 'S*4"...."' )
        self.assertIsInstance( m, re.Match )

    def test_macrodef( self ):
        m = regs.c.macrodef.match( '<=abc>LLHB</abc>' )
        self.assertIsInstance( m, re.Match )

    def test_macrouse( self ):
        m = regs.c.macrouse.match( '<abc />' )
        self.assertIsInstance( m, re.Match )

    def test_macroimport( self ):
        matches = [
            '<+filename/>',
            '<+filename></+>',
            '<+filename ns></+>',
            '<+"filename with space" ns/>',
        ]

        for match in matches:
            m = regs.c.macroimport.match( match )
            self.assertIsInstance( m, re.Match, msg = 'did not match "%s"' % match )

    def test_repeat( self ):
        m = regs.c.repeat.match( '*3' )
        self.assertIsInstance( m, re.Match )

class test_hexer( unittest.TestCase ):
    def setUp( self ):
        # test file content
        tf = b'\0\0\0\1\0\x0a\0\x0bABCD"FGH\0\2\0\3'
        self.data_file = io.BytesIO()
        self.data_file.write( tf )
        self.data_file.seek( 0 )

        self.hx_file = io.StringIO()
        self.out_file = io.StringIO()

        self.hx = hexer.hexer( self.data_file, self.hx_file, self.out_file )

    def hx_write( self, hx ):
        self.hx_file.write( hx )
        self.hx_file.seek( 0 )

    def assertOut( self, out ):
        self.out_file.seek( 0 )
        self.assertEqual( self.out_file.read(), out )

    # __init__
    # run

    def test_run_to_end( self ):
        self.hx_file.write( '@>LHHS*8HH' )
        self.hx_file.seek( 0 )
        self.hx.run()
        self.out_file.seek( 0 )
        self.assertEqual( self.out_file.read(), '@>LHH[1,10,11]S*8"ABCD\\"FGH"HH[2,3]' )

    # get_tokens

    def test_multiline_str( self ):
        self.hx_write( ': S*4"A\nB\n"\n' )
        self.hx.run()
        self.assertOut( '00: S*4"\\x00\\x00\\x00\\x01"\n04: 000a 000b 4142 4344 # <....ABCD>\n0c: 2246 4748 0002 0003 # <"FGH....>\n' )

    def test_repeat( self ):
        self.hx_write( '@> H*10' )
        self.hx.debug = True
        self.hx.run()
        self.assertOut( '@> H[0]H[1]H[10]H[11]H[16706]H[17220]H[8774]H[18248]H[2]H[3]' )

    def test_bad_token( self ):
        self.hx_write( '~~~' )
        with self.assertRaises( Exception ) as cx:
            self.hx.run()
        self.assertEqual( str( cx.exception ), 'could not match: ~~~' )

    def test_str_no_end( self ):
        self.hx_write( 'S*8"\n' )
        with self.assertRaises( hexer.NoEndPastEOF ) as cx:
            self.hx.run()
        self.assertEqual( str( cx.exception ), "no end past eof: 'S*8\"\\n'" )

    @unittest.skip( "to be fixed in v0.1.0" )
    def test_repeat_next_line( self ):
        self.hx_write( ': ....\n*2\n' )
        self.hx.run()
        self.assertOut( '00: 0000\n0001\n04: 000a 000b 4142 4344 # <....ABCD>\n0c: 2246 4748 0002 0003 # <"FGH....>\n' )

    def test_interpolate_str( self ):
        self.hx._vars['test'] = '123'
        self.assertEqual( self.hx.interpolate_str( r'something{test}' ), 'something123' )

    def test_interpolate_name( self ):
        self.hx._vars['test'] = '123'
        self.hx._vars['ns.test'] = '321'
        self.assertEqual( self.hx.interpolate_name( r'name_{test}' ), 'name_123' )
        self.hx.namespace = [ 'ns' ]
        self.assertEqual( self.hx.interpolate_name( r'name_{test}', True ), 'ns.name_321' )

    def test_parse_str( self ):
        self.assertEqual( self.hx.parse_str( r'"something"' ), 'something' )
        self.assertEqual( self.hx.parse_str( r'"escaped\""' ), 'escaped\\"' )

    def test_get_vars( self ):
        self.assertEqual( self.hx.get_vars( r'{a}' ), [ 'a' ] )
        self.assertEqual( self.hx.get_vars( r'{a,b,c}' ), [ 'a', 'b', 'c' ] )
        self.hx.namespace = [ 'ns' ]
        self.hx._vars['ns.i'] = 0
        self.assertEqual( self.hx.get_vars( r'{a_{i}}' ), [ 'ns.a_0' ] )

    def test_get_var( self ):
        self.hx._vars['a'] = 'something'
        self.hx._vars['a_something'] = 'else'
        self.assertEqual( self.hx.get_var( r'a' ), 'something' )
        self.assertEqual( self.hx.get_var( r'a_{a}' ), 'else' )

    def test_set_var( self ):
        self.hx.set_var( r'a', 'something' )
        self.assertEqual( self.hx._vars['a'], 'something' )

    def test_get_val( self ):
        self.hx._vars['a'] = 'something'
        self.assertEqual( self.hx.get_val( r'{a}' ), 'something' )
        self.hx._vars['a'] = 10
        self.assertEqual( self.hx.get_val( r'{a}' ), 10 )

class test_main( unittest.TestCase ):
    def setUp( self ):
        # test file content
        tf = b'\0\0\0\1\0\x0a\0\x0bABCD"FGH\0\2\0\3'

        self.temp_dir = tempfile.TemporaryDirectory()
        self.data_file = pathlib.Path( self.temp_dir.name, 'data_file' )
        self.hx_file = pathlib.Path( self.temp_dir.name, 'data_file.hx' )

        self.data_file.write_bytes( tf )

    def tearDown( self ):
        self.temp_dir.cleanup()

    def test_no_args( self ):
        with self.assertRaises( SystemExit ) as cm:
            main( [] )

        self.assertEqual( cm.exception.code, 2 )

    def test_no_hx( self ):
        main( [ str( self.data_file ) ] )

        self.assertTrue( self.hx_file.exists() )
        self.assertEqual( self.hx_file.read_text(), '00: 0000 0001 000a 000b # <........>\n08: 4142 4344 2246 4748 # <ABCD"FGH>\n10: 0002 0003 ' )

    def test_hx( self ):
        self.hx_file.write_text( '00: @>LHH\n' )

        main( [ str( self.data_file ) ] )

        self.assertTrue( self.hx_file.exists() )
        self.assertEqual( self.hx_file.read_text(), '00: @>LHH[1,10,11]\n08: 4142 4344 2246 4748 # <ABCD"FGH>\n10: 0002 0003 ' )

    def test_json( self ):
        self.hx_file.write_text( '00: @>S{a}*8\n' )

        main( [ str( self.data_file ), '-j' ] )

        self.assertTrue( self.hx_file.exists() )
        self.assertEqual( self.hx_file.read_text(), '00: @>S{a}*8"\\x00\\x00\\x00\\x01\\x00\\n\\x00\\x0b"\n08: 4142 4344 2246 4748 # <ABCD"FGH>\n10: 0002 0003 ' )

    def test_debug( self ):
        self.hx_file.write_text( '00: @> H*10\n' )

        main( [ str( self.data_file ), '-d' ] )

        self.assertTrue( self.hx_file.exists() )
        self.assertEqual( self.hx_file.read_text(), '00: @> H[0]H[1]H[10]H[11]H[16706]H[17220]H[8774]H[18248]H[2]H[3]\n' )

    def test_default( self ):
        main( [ str( self.data_file ), '-m', ': . . . .' ] )

        self.assertTrue( self.hx_file.exists() )
        self.assertEqual( self.hx_file.read_text(), '00: 00 00 00 01\n04: 00 0a 00 0b\n08: 41 42 43 44\n0c: 22 46 47 48\n10: 00 02 00 03\n' )
        
if __name__ == '__main__': # pragma: no cover
    unittest.main()



import re

class c:
    def __setattr__( self, k, v ):
        self.__dict__[ k ] = v if isinstance( v, re.Pattern ) else re.compile( v )

c = c()

comment = r'(?:#( \<.*\>)?(.*))'
c.comment = re.compile( comment + '$', re.MULTILINE )

c.space = space = r'\s+'

c.filepos = re.compile( r'(?P<start>^\*?---\s*)?(?P<rel>\+?)(?P<pos>[a-f0-9]*)(?P<end>:)', re.MULTILINE )

c.hex = r'(=?[a-f0-9.]+)'

name = r'[\w.\-{}!?]+'

var = r'\{\s*' + name + r'\s*\}'
_vars = r'\{\s*' + name + r'\s*(?:,\s*' + name + r')*\s*\}'
values = r'(?:[0-9\.]+|' + var + r')'

c.endian = r'@[<>@]'

c.dtypes = dtypes = r'(?:\-?[OBIHLEFD])'
c.types = r'(' + dtypes + r'+)(' + _vars + r')?(?:\[([^\]]+)\])?'

string_inner = r'(?:\\"|[^"])*'
string = r'(?P<start>")(' + string_inner + r')(?P<end>")?'
c.string = r'(?P<type>[SU](?:|0))(?P<name>' + var + r')?(?:\*(?P<size>' + values + r'))?(?:' + string + r')?'

c.macrodef = re.compile( r'(?P<start><=(?P<name>' + name + r')>)((?:[^<]|<(?!\/(?P=name)>))*)(?P<end><\/(?P=name)>)?', re.DOTALL )
c.macrouse = re.compile( r'<(?P<name>' + name + r')\s*(?P<ns>' + name + r')?(?:(?P<start>>)((?:[^<]|<(?!\/(?P=name)>))*)(?P<end><\/(?P=name)>)?|\s*?(?P<short_end>\/>))', re.DOTALL )
c.macroimport = re.compile( r'<\+(?P<filename>' + name + r'|"' + string_inner + r'")\s*(?P<ns>' + name + r')?(?:(?P<start>>)((?:[^<]|<(?!\/\+>))*)(?P<end><\/\+>)?|\s*?(?P<short_end>\/>))', re.DOTALL )

repeat = r'(?:(\s*)\*\s*(?P<repeat>' + values + r'))'
c.repeat = repeat + r'+'

if __name__ == '__main__': # pragma: no cover
    for k, v in dict( vars() ).items():
        print( k, ':', sep = '' )
        if k == 'c':
            for ck, cv in vars( c ).items():
                print( '\t', ck, ':', sep = '' )
                print( '\t', cv.pattern )
            continue
        print( v )



#!/bin/env python
import argparse
import io
import math
import os.path
import pathlib
import re
import struct
import sys

from . import regs

class FilePosMissmatch( RuntimeError ):
    pass

class HexDataAssertError( AssertionError ):
    pass

class NoEndPastEOF( SyntaxError ):
    pass

class InvalidToken( SyntaxError ):
    pass

class MacroNotFound( SyntaxError ):
    pass

# defaults
default_macro = ': .... .... .... .... # <>\n'

def typesr_to_struct( m ):
    '''
    converts a typesr re.match obj to a struct.Struct obj
    '''
    return ''.join( [
        ( t.startswith( '-' ) and t[1:].lower() )
        or ( t in [ 'F', 'D', 'E' ] and t.lower() )
        or t
        for t in regs.c.dtypes.findall( m[0] )
    ] )

def filepos_token( m, hx ):
    skip = m['start']
    trel = bool( m['rel'] )
    fp = hx.data_file.tell()
    if m['pos']:
        tfp = int( m['pos'], 16 )
    else:
        tfp = fp

    if trel:
        tfp = fp + tfp

    if skip or trel:
        # skipping
        if skip and skip.startswith( '*' ):
            raise NotImplementedError( 'output skipped data not implemented' )
        else:
            hx.data_file.seek( tfp )
        fp = tfp
    else:
        # ensure correct filepos is given
        if tfp != fp:
            raise FilePosMissmatch( "file position missmatch {0} vs {1}".format( hex( fp ), hex( tfp ) ) )

    yield ( skip or '' ) + hex( fp )[2:].rjust( hx.fpw, '0' ) + ':'

def comment_token( m, hx ):
    # insert/replace readable chunk
    if m[1]:
        yield r'# <%s>' + m[2]
    else:
        yield r'#' + m[2]

def hex_token( m, hx ):
    hex_data = m[1]
    assert_data = hex_data.startswith( '=' )

    if assert_data:
        yield '='
        hex_data = hex_data[1:]

    hex_len = math.ceil( len( hex_data ) / 2 )
    chunk = hx.data_file.read( hex_len )

    if not chunk:
        raise EOFError

    chunk = chunk.hex()

    if assert_data and hex_data != chunk:
        raise HexDataAssertError( '%s != %s' % ( hex_data, chunk ) )

    yield chunk

def endian_token( m, hx ):
    hx.endianess = m[0][1]
    yield m[0]

def types_token( m, hx ):
    # get struct def
    s = typesr_to_struct( m )

    # get vars to store in (if any)
    vs = hx.get_vars( m[2] ) if m[2] else None

    s = struct.Struct( ( hx.endianess or '' ) + s )
    vals = s.unpack( hx.data_file.read( s.size ) )
    if vs:
        hx._vars.update( zip( vs, vals ) )
    yield m[1] + ( m[2] or '' ) + "[{0}]".format( ','.join( [ str( v ) for v in vals ] ) )

def string_token( m, hx ):
    data = hx.data_file.read( hx.get_val( m['size'] ) ).rstrip( b'\0' )
    if m['name']:
        for v in hx.get_vars( m['name'] ):
            hx._vars[ v ] = data
    data = str( bytes( str( data, encoding = 'ascii', errors="backslashreplace" ), encoding = 'unicode_escape' ), encoding = 'ascii' ).replace( '"', '\\"' )
    yield m['type'] + \
        ( m['name'] or '' ) + \
        ( m['size'] and ( '*' + m['size'] ) or '' ) + \
        '"%s"' % data

def skip( m, hx ): # pragma: no cover
    yield '' 

def ret( m, hx ): # pragma: no cover
    yield m[0]

def macro_def_token( m, hx ):
    hx.macros[ hx.interpolate_name( m['name'], False ) ] = m[3]

    yield m[0]

def macro_use_token( m, hx ):
    name = hx.interpolate_name( m['name'], False )
    ns = m['ns'] and hx.interpolate_name( m['ns'], False ) or ''

    yield '<%s>' % ( name + ( ns and ( ' ' + ns ) ) )

    try:
        yield from hx.get_tokens( hx.macros[ name ], ns )
    except KeyError:
        raise MacroNotFound( name )

    yield '</%s>' % name

def macro_import_token( m, hx ):
    ns = m['ns'] and hx.interpolate_name( m['ns'], False ) or ''
    short_end = m['short_end']

    filename = hx.interpolate_name( m['filename'], False )
    if filename.startswith( '"' ):
        filename = hx.parse_str( filename )

    yield '<+%s>' % ( m['filename'] + ( ns and ( ' ' + ns ) ) + ( short_end and '/' or '' ) )

    out = io.StringIO() if not short_end else None

    hexer(
        hx.data_file, pathlib.Path( filename ).resolve().open(), out,
        _vars = hx._vars, macros = hx.macros, namespace = hx.namespace,
        run_default = False, debug = hx.debug
    ).run()

    if out:
        yield out.getvalue()

    if not short_end:
        yield '</+>'

def repeat_token( m, hx, token, recurse = None ):
    th, tm = token

    if not recurse:
        recurse = list( re.finditer( regs.repeat, m[0] ) )

    m = recurse[-1]
    repeats = hx.get_val( m['repeat'] )

    for i in range( repeats ):
        hx.set_var( 'i', i )
        if len( recurse ) > 1:
            yield from repeat_token( m, hx, token, recurse[:-1] )
        else:
            yield from th( tm, hx )

        if m[1]:
            yield m[1]

    if hx.has_var( 'i' ):
        hx.del_var( 'i' )

class hexer:
    tokens = [
        ( regs.c.endian, endian_token, 'ndi' ),
        ( regs.c.filepos, filepos_token, 'fp' ),
        ( regs.c.comment, comment_token, '#' ),
        ( regs.c.hex, hex_token, 'hx' ),
        ( regs.c.types, types_token, 't' ),
        ( regs.c.string, string_token, '"' ),
        ( regs.c.macrodef, macro_def_token, '=mac' ),
        ( regs.c.macrouse, macro_use_token, 'mac' ),
        ( regs.c.macroimport, macro_import_token, '+mac' ),
        ( regs.c.space, ret, '_' ),
    ]

    def __init__( self, data_file, hx_file = None, out_file = None, _vars = None, macros = None, namespace = None, run_default = True, debug = False ):
        global default_macro

        self.data_file = data_file
        self.hx_file = hx_file
        self.out_file = out_file

        self.debug = debug

        # context vars
        self._vars = _vars or {}
        self.macros = macros or {}

        self.run_default = run_default

        if 'default' not in self.macros and self.run_default:
            self.macros['default'] = default_macro

        self.namespace = namespace or [] # current namespace

        self.prev_token = None

        self.data_size = -1

        if hasattr( self.data_file, 'name' ):
            self.data_size = os.path.getsize( self.data_file.name )
        elif isinstance( self.data_file, io.BytesIO ):
            self.data_size = len( self.data_file.getbuffer() )

        self.fpw = 0

        if self.data_size != -1:
            # work out max filepos width
            self.fpw = len( hex( self.data_size )[2:] )

        self.eof = False
        self.endianess = None
        self.skipping = False

    def run( self ):
        while not self.eof:
            pos = self.data_file.tell()

            if pos == self.data_size:
                break

            out_chunk = ''
            readable_chunk = b''

            if self.hx_file:
                hchunk = self.hx_file.readline()

                if not hchunk:
                    self.hx_file.close()
                    self.hx_file = None
                    continue
            elif self.macros['default'] and self.run_default:
                hchunk = self.macros['default']
            else:
                break

            f_pre_tok = self.data_file.tell()

            try:
                for token in self.get_tokens( hchunk ):
                    out_chunk += token
            except EOFError:
                self.eof = True

            f_post_tok = self.data_file.tell()

            self.data_file.seek( f_pre_tok )

            readable_chunk = re.sub( b'[^ -~]', b'.', self.data_file.read( f_post_tok - f_pre_tok ) )

            out_chunk = out_chunk.replace( r'# <%s>', readable_chunk and ( '# <' + str( readable_chunk, encoding = 'ascii' ) + '>' ) or '#' )

            if self.out_file:
                print( out_chunk, file = self.out_file, end = '' )

    def interpolate_str( self, v, ns = True ):
        '''
        something{var} -> something123
        '''
        m = re.search( regs.var, v )
        while m:
            v = re.sub( regs.var, str( self.get_val( m[0] ) ), v )
            m = re.search( regs.var, v )

        return v

    def interpolate_name( self, v, ns = True ):
        '''
        name_{i} -> name_1
        v -> ns.v
        '''
        v = self.interpolate_str( v )

        if ns:
            return '.'.join( self.namespace + [ v ] )

        return v

    def parse_str( self, v ):
        '''
        '"something"'' -> 'something'
        '"escaped\""' -> 'escaped"'
        '''
        return v.replace( '\"', '"' )[1:-1]

    def get_vars( self, v ):
        '''
        {a} -> [ 'a' ]
        {a,b,c} -> [ 'a', 'b', 'c' ]
        {a_{i}} -> [ 'ns.a_0' ]
        '''
        return [ self.interpolate_name( v ) for v in v[ 1: -1 ].split( ',' ) ]

    def get_var( self, v ):
        return self._vars[ self.interpolate_name( v ) ]

    def set_var( self, v, val ):
        self._vars[ self.interpolate_name( v ) ] = val

    def has_var( self, v ):
        return self.interpolate_name( v ) in self._vars

    def del_var( self, v ):
        del self._vars[ self.interpolate_name( v ) ]

    def get_val( self, v ):
        return self._vars[ self.get_vars( v )[0] ] if v.startswith( '{' ) else int( v )

    def get_tokens( self, data, ns = '' ):
        if ns:
            self.namespace.append( ns )

        while data:
            for r, handler, name in self.tokens:
                m = r.match( data )

                if m:
                    if self.debug:
                        print( name, repr( m[0] ), m.groups(), m.groupdict(), file = sys.stderr )

                    while 'start' in m.groupdict() and m['start'] and not m['end']:
                        # token started but not ended in data
                        # keep reading lines from hx
                        # until token end found
                        chunk = self.hx_file.readline()
                        if chunk == '':
                            raise NoEndPastEOF( "no end past eof: " + repr( data[:20] ) )
                        data += chunk
                        m = r.match( data )
                    
                    data = data[ m.end() : ]

                    # is next token repeat?
                    rm = regs.c.repeat.match( data )
                    if rm:
                        yield from repeat_token( rm, self, ( handler, m ) )
                        data = data[ rm.end() : ]
                    else:
                        yield from handler( m, self )


                    if r != regs.c.space:
                        self.prev_token = m[0]

                    break
            else:
                raise InvalidToken( "could not match: {0}".format( data ) )

        if ns:
            self.namespace.pop()


def main( argv = None ):

    parser = argparse.ArgumentParser(
        formatter_class = argparse.RawDescriptionHelpFormatter,
        description = '''
    when file.hx does not exist
        read file and output file.hx (via file.hx.out)

    when file.hx exists
        read file and parse file.hx to output file.hx (via file.hx.out)
''',
        epilog = '''
syntax:
    see https://gitlab.com/hexer-py/hexer-docs/-/blob/master/syntax.md
''' )

    # pos
    parser.add_argument( 'file', help = 'binary file to decode' )
    parser.add_argument( 'file.hx', nargs = '?', help = 'hexer file' )

    # opt
    parser.add_argument( '-o', '--output', metavar = 'file.hx.out', nargs = '?', help = 'sets output file to file.hx.out (default)' )
    parser.add_argument( '-k', '--keep-output', action = 'store_true', help = 'output file will be kept as is, instead of moved to file.hx' )
    parser.add_argument( '-w', '--write', action = 'store_const', const = 'write', dest = 'mode', help = 'use file.hx to write content to file (**not implemented**)' )
    parser.add_argument( '-v', '--verify', action = 'store_const', const = 'verify', dest = 'mode', help = 'use file.hx to verify content of file (**not implemented**)' )
    parser.add_argument( '-d', '--debug', action = 'store_true', help = 'tokenizer will output debug info to stderr' )
    parser.add_argument( '-j', '--json', action = 'store_true', help = 'output vars as json to stdout' )
    parser.add_argument( '-m', '--default', metavar = 'macro', help = 'set default macro (used to create or fill out file.hx)' )

    args = parser.parse_args( argv )
    
    file = args.file
    file_hx = vars( args )['file.hx'] or file + '.hx'
    file_hx_out = args.output or file_hx + '.out'

    with open( file, 'rb' ) as f, open( file_hx_out, 'w' ) as out:
        fhx = open( file_hx, 'r' ) if os.path.exists( file_hx ) else None
    
        h = hexer( f, fhx, out, debug = args.debug )

        if args.default:
            h.macros['default'] = args.default + '\n'

        h.run()

        if fhx:
            fhx.close()

        if args.json:
            import json

            def encode_bytes( obj ):
                if isinstance( obj, bytes ):
                    return str( obj, encoding = 'ascii' )

            print( json.dumps( h._vars, default = encode_bytes ) )

    if not args.keep_output:
        os.rename( file_hx_out, file_hx )

if __name__ == '__main__': # pragma: no cover
    main( sys.argv )



#!/usr/bin/env python
from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name = 'hexer-hx',
    version = '0.0.2',
    author = 'Martin Hultén-Ashauer',
    author_email = 'hexer-info@nimdraug.com',
    description = 'a tool to help in the arcane arts of decoding and reverse engineering binary data',
    long_description = long_description,
    long_description_content_type="text/markdown",
    url = 'https://gitlab.com/hexer-py/hexer',
    license = 'MIT',

    packages = [ 'hexer' ],

    classifiers = [
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "Intended Audience :: System Administrators",
        "Intended Audience :: Telecommunications Industry",
        "Topic :: Text Processing :: Markup",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Utilities",
    ],

    entry_points = {
        'console_scripts': [
            'hexer=hexer:main',
        ],
    },
)
